# Copyright (C) 2006 Ulrik Sverdrup
#
#   dragbox -- Commandline drag-and-drop tool for GNOME
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

from sys import exit, argv
from os import path

from utils import print_error

class dragbox(object):
	u"""
	dragbox controller object
	
	"""
	
	def __init__(self):
		""" 
		This class runs one instance of dragbox
		+ Read commandline
		+ Try dbus connection
		+ Add items
		+ Run 
		""" 
		# Read commandline switches
		(files, texts, cli_flags) = self.read_opts(argv[1:])

		# Read from stdin
		from sys import stdin
		if not stdin.isatty():
			pipe_text = stdin.read()
			if pipe_text:
				texts += [pipe_text]
				
		nofork = cli_flags.get("no-fork") # there is a cli flag
		if not nofork:
			sent_to_dbus = self.try_dbus_send(files,texts)
			if sent_to_dbus:
				exit(0) # Done, sent to other running instance
			# Fork!
			from os import fork
			pid = fork()
			if pid:
				exit(0)
				
		from window import window_controller
		self.wc = window_controller(cli_flags)
		self.add_files(files)
		self.add_texts(texts)
		
		try:
			self.wc.run_window()
		except KeyboardInterrupt:
			# catch ctrl-c and fail nicely
			exit(0)
	def read_opts(self, argv):
		""" 
		Uses gnu getopt to read command line arguments
		returns (files, texts, options) 
		
		""" 
		from getopt import gnu_getopt, GetoptError
		try:
			(options, arguments) = gnu_getopt(argv, "nhf:t:vqpu0", ['help', 'file=', 'text=', 'no-fork', 'version', 'quiet' , 'paths', 'uris', 'null-terminate'])
		except GetoptError, exception:
			print_error("Sorry, %s" % exception) # prints "option * not recognized"
			self.print_usage()
		prefs = {}
		# Search for preferences
		for (o,a) in options: 
			if o == "-h" or o == "--help":
				self.print_usage()
			if o == "-v" or o == "--version":
				self.print_version()
			if o == "-n" or o == "--no-fork":
				prefs["no-fork"] = True
				if not prefs.get("quiet"): # make sure we don't ow
					prefs["quiet"] = False
			if o == "-q" or o == "--quiet":
				prefs["quiet"] = True
			if o in ("-p", "--paths"):
				prefs["paths"] = True
			if o in ("-u", "--uris"):
				prefs["paths"] = False
			if o in ("-0", "--null-terminate"):
				prefs["null"] = True
		# Handle strongly typed items
		files = []
		texts = []
		for (o, a) in options:
			if o == "-f" or o == "--file":
				path = self.check_file_exists(a, warn=True)
				if path: files.append(path) 
			elif o == "-t" or o == "--text":
				texts.append(a)
		# Handle rest
		# Try if they are files
		text = ""
		for item in arguments :
			path = self.check_file_exists(item, warn=False)
			if path:
				files.append(path)
			else:
				text = text + " " + item
				
		# Take rest as one text block
		if text: texts.append(text)
		return (files, texts, prefs)
	def init_gnome(self): 
		""" 
		we have to init gnome properly
		"""  
		import globals
		from gnome import init, PARAM_APP_DATADIR, PARAM_ENABLE_SOUND
		properties = { 
				PARAM_APP_DATADIR : globals.data_dir,
				PARAM_ENABLE_SOUND : False 
			}
		init(u"dragbox", globals.version, properties=properties)
		
	def try_dbus_send(self, files, texts): 
		""" 
		Tries to connect ot already running instance, 
		if that fails starts a dbus service 
		 
		""" 
		
		from globals import dbus_enabled
		if dbus_enabled != "yes":
			return False #return with no success
			
		import dbuscontrol
		from utils import print_debug 
		should_die = False
		
		try:
			dclient = dbuscontrol.client()
			iface = dclient.get_interface()
			#print "Connected to dbus server"
			if files: iface.send_files(files)
			if texts: iface.send_texts(texts)
			should_die = True
			print_debug("Sent files, quitting")
		except:
			print_debug("No dbus service found, starting one")
			try:
				dserver = dbuscontrol.make_server()
				dserver.set_delegate(self)
			except:
				print_error("Failed to start dbus, might not be available")
 		return should_die
	def check_file_exists(self, f, warn=True):
		""" 
		Checks if a file exists
		@param f: the local path to the file 
		@type f: a string object. 
		
		@param warn: If True, this method prints an error message and exits 
		@type warn: a bool object.    
		""" 
		p = path.abspath(f)
		if path.exists(p):
			return p
		if not warn:
			return None
		print_error("The file %s does not exist" % f)
		exit(1)
	def add_files(self, flist):
		""" 
		Adds a list of files to the dragbox
		@param flist: List of absolute paths  
		@type flist: a list object.   
		 
		""" 
		from dragitem import dragitem, FILE_TYPE
		 
		for f in flist:
			p = path.abspath(f)
			d = dragitem(p, FILE_TYPE)
		 	self.wc.add_dragitem(d)
	def add_texts(self, tlist):
		""" 
		Adds a list of text to the dragbox
		@param tlist: List of strings 
		@type tlist: a list object.   
		 
		""" 
		from dragitem import dragitem, TEXT_TYPE
		 
		for t in tlist:
			d = dragitem(t, TEXT_TYPE)
			self.wc.add_dragitem(d)
	def activate_application(self):
		""" 
		Activates, brings to front
		For dbus use 
		 
		""" 
		self.wc.show_window() #make sure it's in front
		
	def print_usage(self): 
		print """
Usage: %s [-nq] item [...]

    item                    add item to dragbox
                            added as file if it is a valid path
                        	
    -t, --text "text"       add text
    -f, --file file         add file
	
    -n, --no-fork           don't fork (also disables use of dbus)
    
    The options below only apply with --no-fork:
    -q, --quiet             don't print dragged-in items
    -p, --paths             print paths of dragged-in files (default)
    -u, --uris              print uris of dragged-in files
    -0, --null-terminate    separate items' output with \\0, not \\n 
                      
	
    -h, --help              display this help
    -v, --version           display version information
		""" % "dragbox"
		exit(1)
	def print_version(self):
		import globals
		print "%s %s\n\n%s" % (globals.package_name, globals.version, globals.copyright_info)
		exit(0)
		