# Copyright (C) 2006 Ulrik Sverdrup
#
#   dragbox -- Commandline drag-and-drop tool for GNOME
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA 

import pygtk
pygtk.require('2.0')
import gtk

from dragitem import dragitem, TEXT_TYPE, FILE_TYPE, TARGET_TYPE_TEXT, TARGET_TYPE_URI_LIST
import preferences
from utils import print_error, print_debug

class window_controller(object):
	u"""
	Dragbox' window controller object
	Keeps track of a window, a list of dragitem
	
	Handles drag-and-drop and copy and paste
	Handles pop-up menus
	"""
	def __init__(self, cli_flags):
		""" 
		__init__ for window_controller
		
		@param cli_flags: CLI options 
		@type cli_flags: a dictionary object.   
		""" 
		self.dragitems = {}
		self.cli_flags = cli_flags
		self.placeholder = None
		
		self.create_window()
		
		preferences.notify_add("stick", self.cb_stick_pref_changed)
	
	def show_window(self):
		"""
		Moves the window to front
		""" 
		already_sticks = self.window.stick_initially
		if not already_sticks:
			self.window.stick()
		self.window.deiconify()
		self.window.present()
		if not already_sticks:
			self.window.unstick()
		
		
	def run_window(self): 
		""" 
		Shows the controlled window,
		sets window properties and starts the
		GTK main loop
		 
		""" 
        # show it all
		self.window.show_all()
		self.adjust_window_size()		
		# Check the sticky preference
		stick = preferences.get_bool("stick", if_none=True)
		if stick: 
			self.window.stick()		
		
		gtk.main()
				 
		
	def win_delete_event(self, event, more):
		""" 
		window close callback 
		 
		""" 
		return False
	
	def win_destroy(self, event): 
		""" 
		window destroy callback
		terminates the GTK main loop 
		
		""" 
		from utils import quit_dragbox 
		quit_dragbox()
	
	def create_window(self): 
		"""
		create a new window 
		"""
		self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.window.set_title("Dragbox")
    
		self.window.connect("delete_event", self.win_delete_event)
		self.window.connect("destroy", self.win_destroy)
		self.window.connect("button-press-event", self.cb_window_click)

		self.vbox = gtk.VBox(False, 0)

		# Add a scrollbox around the vbox
		# scrolls the list when it's too long
		scrollbox = gtk.ScrolledWindow()
		scrollbox.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
		scrollbox.set_border_width(0)
		self.window.add(scrollbox)
		scrollbox.add_with_viewport(self.vbox)
		
		# Add placeholder for newbies
		self.make_placeholder() 
		
		pixbuf = self.window.render_icon(gtk.STOCK_DND_MULTIPLE, gtk.ICON_SIZE_DIALOG)
		self.window.set_icon(pixbuf)
		
		# Setup window as a drag TARGET
		from dragitem import file_targets, text_targets
		self.window.drag_dest_set(gtk.DEST_DEFAULT_ALL, file_targets+text_targets, gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_COPY )
		self.window.connect("drag-data-received", self.cb_drag_data_received)
	
	def adjust_window_size(self, shrink=False):
		""" 
		Changes the window size to better fit the content 
		
		@param shrink: If the window is allowed to shrink horizontally 
		@type shrink: a bool object.  
		""" 
		#Resize window horizontally
		minwidth = 200
		minheight = 200
		maxheight = 500
		
		win = self.window
		widget = self.vbox.get_parent() #The viewport in the scrollbox
		scrollbox = self.window.get_child()
		(oldw, oldh) = win.get_size()
		(w, h) = widget.size_request()
		
		if w < 10:
			# Empty
			self.make_placeholder()
			return  
		
		w = max(minwidth, w)
		h = max(minheight, h)
		h = min(maxheight, h)
		
		w = max(w, oldw)
		
		if not shrink:
			vadj = scrollbox.get_vadjustment()
			if h > vadj.page_size:
				vadj.value = vadj.upper - vadj.page_size
				scrollbox.set_vadjustment(vadj)
				widget.queue_draw()
		
		#if not shrink:
		if h < oldh: h = oldh

		# keep old width
		win.resize(w,h)
	
	def make_placeholder(self): 
		self.placeholder = gtk.Label("Drag files and text here")
		self.placeholder.set_sensitive(False) 
		self.placeholder.show() 
		self.vbox.add(self.placeholder)
		 
	def add_dragitem(self, dragitem):
		""" 
		Adds a dragitem's widget to the windows
		main container
		@param dragitem: Dragitem to be added 
		@type dragitem: a dragitem object.   
		 
		""" 
		widget = dragitem.get_widget()
		if not widget:
			return
			
		# Remove placeholder if there
		if self.placeholder:
			self.vbox.remove(self.placeholder)
			self.placeholder = None 
				
		# Create a hbox to contain the dragitem and
		# a close button on the side
		hbox = gtk.HBox(False,0)
		hbox.pack_start(widget, True, True, 0)
		
		# close button
		closebtn = gtk.Button("")
		image = gtk.Image()
		image.set_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_MENU)
		image.set_padding(0,0)
		image.set_alignment(0, 0)
		closebtn.set_image(image)
		closebtn.set_relief(gtk.RELIEF_NONE)
		closebtn.connect("clicked", self.cb_close_button, hbox)
		closebtn.set_property("can-focus", False)
		closebtn.set_property("height-request", 22)
		closebtn.set_property("width-request", 22)
		
		alignment = gtk.Alignment()
		alignment.add(closebtn)
		alignment.set(1, 0, 0, 0)
		hbox.pack_start(alignment, False,False,0)
		
		# Add the dragitem & close package to main container
		self.vbox.pack_start(hbox,False,True,0)
		
		widget.show()
		self.dragitems[dragitem] = hbox
		self._add_dragitem_callbacks(dragitem)
		self._setup_drag(dragitem)
		
		self.window.show_all()
		from gobject import timeout_add 
		#self.adjust_window_size()
		timeout_add(50, self.adjust_window_size)
	
	def _add_dragitem_callbacks(self, dragitem):
		""" 
		Adds callbacks for individual dragitems
		Adds 'clicked' and 'button-press-event' 
		""" 
		w = dragitem.get_widget();
		hbox = self.dragitems[dragitem];
		
		w.connect("clicked", self.cb_dragitem_click, dragitem)
		w.connect("button-press-event", self.cb_dragitem_other_click, dragitem)
	
	def cb_drag_data_received(self, widget, drag_context, x, y, selection_data, info, timestamp):
		""" 
		Callback when something is dropped on the window 
		""" 
		wid = drag_context.get_source_widget()
		#None if from other application
		if wid:
			return False
			
		if info == TARGET_TYPE_TEXT:
			self._get_text_from_selection(selection_data)
		elif info == TARGET_TYPE_URI_LIST: 
			self._get_files_from_selection(selection_data)
		else:
			print_error("what?")
			return
		drag_context.finish(True, False, timestamp)  #Finished w succ
	#### There is too much duplication below. This has to be refactored
	def _get_text_from_selection(self, selection_data):
		""" 
		internal method to handle dropped text 
		
		@param selection_data: Contains the dropped data 
		@type selection_data: a gtk.SelectionData object.  
		""" 
		text = selection_data.get_text()
		if text:
			d = dragitem(text, TEXT_TYPE)
			self.add_dragitem(d)
			c = self.cli_flags.get("quiet", True)
			null = self.cli_flags.get("null", False)
			if not c:
				d.print_to_stdout(null_terminate=null)
			return True
		return None
	def _get_files_from_selection(self, selection_data):
		""" 
		internal method to handle dropped files 
		
		@param selection_data: Contains the dropped data 
		@type selection_data: a gtk.SelectionData object.  
		""" 
		files = selection_data.get_uris()
		from gnomevfs import get_local_path_from_uri
		
		for f in files:
			try:
				p = get_local_path_from_uri(f)
				type = FILE_TYPE
			except:
				p = f
				type = TEXT_TYPE	
			if p:
				d = dragitem(p, type)
				self.add_dragitem(d)
				c = self.cli_flags.get("quiet", True)
				paths = self.cli_flags.get("paths", True)
				null = self.cli_flags.get("null", False)
				if not c:
					d.print_to_stdout(paths, null_terminate=null)
		return True
	def _setup_drag(self, dragitem):
		""" 
		Method to setup a widget as a drag source 
		widget may be a dragitem widget
		 
		""" 
		w = dragitem.get_widget()
		user_info = dragitem

		w.drag_source_set(gtk.gdk.BUTTON1_MASK,
				dragitem.get_targets(),
				gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE | gtk.gdk.ACTION_LINK)
		w.connect("drag-data-get", self.cb_drag_send, user_info)
		w.connect("drag-end", self.cb_drag_end, user_info)
	def cb_drag_end(self, widget, context, dragitem):
		""" 
		Callback when drags end 
		 
		""" 
		if context.dest_window:	
			# Successful drag
			# check if we should remove anyways
			remove = preferences.get_bool("remove", False)
			if remove:
				self._remove_dragitem(dragitem)
				return
			
			# Remove if it's a move action and it moved
			# We have to wait for the filesystem to react
			# Make one quick and one slow to catch it all!
			if context.action is gtk.gdk.ACTION_MOVE and dragitem.get_type() == FILE_TYPE:
				from gobject import timeout_add
				# Wait 0.1s and 3.0s
				timeout_add(100, self._remove_dragitem_if_gone, dragitem)
				timeout_add(3000, self._remove_dragitem_if_gone, dragitem)
	def cb_drag_send(self, widget, context, selection, target_type, eventTime, dragitem):
		""" 
		Callback
		Called when source wants the data 
		 
		""" 
		dragitem.write_to_selection(selection, target_type)
		return True
	def cb_close_button(self, widget, widgetbox): 
		""" 
		callback from close button
		takes the close widget and the hbox to be removed 
		 
		""" 
		self._remove_child(widgetbox)
	def cb_dragitem_click(self, widget, dragitem): 
		""" 
		Callback from button click 
		copies to clipboard or opens file
		"""  
		if dragitem.get_type() is FILE_TYPE:
			open = preferences.get_bool("open", if_none=False)
			if open:
				self._open_file(dragitem)
				return
		self._copy_to_clipboard(dragitem)
	def _copy_to_clipboard(self, dragitem): 
		""" 
		copies the represented object of arg to
		clipboard 
		""" 
		clip = gtk.clipboard_get() # get default clipboard
		owner = clip.get_owner() 
		if owner is not None:
			return 
		
		drag_targets = dragitem.get_targets()
		clip.set_with_data(drag_targets, self.cb_clipboard_copy, self.cb_clipboard_done, user_data=dragitem) 
		clip.set_can_store(drag_targets)
		clip.store() # force immediate retrieval?

	def _open_file(self, dragitem): 
		""" 
		opens the file of the dragitem argument 
		""" 
		#from utils import run_process
		#file_path = dragitem.get_object()
		#run_process("gnome-open", [file_path])
		from gnome import url_show 
		file_uri = dragitem.get_uri()
		try:
			url_show(file_uri)
		except:
			from utils import print_error
			print_error("Could not open %s" % file_uri) 
			
	def cb_clipboard_copy(self, clipboard, selectiondata, target_type, user_data=None):
		""" 
		callback called when clipboard copy data is retrieved 
		 
		""" 
		
		dragitem = user_data
		if dragitem:
			dragitem.write_to_selection(selectiondata, target_type)
		 
		return True
	def cb_clipboard_done(self, clipboard, user_data=None):
		""" 
		callback when clipboard is relaxed
		currently does nothing 
		""" 
		#Do nothing
		pass
		
	def cb_dragitem_other_click(self, widget, event, dragitem):
		""" 
		callback when the mouse button is clicked on the window
		handles mouse buttons 2 and 3 
		""" 
		if event.button is 3:
			# Right click
			menu = make_menu(self, dragitem)
			wid = dragitem.get_widget()
			wid.grab_focus()
			menu.popup(None,None,None, event.button, event.time)
			return True
		return False # Propagate event
	def cb_window_click(self, widget, event, user_info=None): 
		""" 
		Callback from button click in window
		handles popup menu  
		""" 
		if event.button is 3:
			# Right click
			menu = make_window_menu(self, None)
			#print menu
			menu.popup(None,None,None, event.button, event.time)
			return True
		if event.button is 2:
			(x,y) = event.get_root_coords()
			self.window.begin_move_drag(event.button, int(x), int(y), event.time)
			return True
		return False
	def cb_menu_copy(self, dragitem): 
		""" 
		callback from context menu
		copies the selected dragitem's representation
		""" 
		self._copy_to_clipboard(dragitem)
	def cb_menu_cut(self, dragitem): 
		""" 
		callback from context menu
		copies the selected dragitem's representation
		""" 
		self._copy_to_clipboard(dragitem)
		self._remove_dragitem(dragitem)
	def cb_menu_remove(self, dragitem): 
		""" 
		callback from context menu
		Removes the selected dragitem 
		""" 
		vbox = self.dragitems.get(dragitem)
		self._remove_child(vbox)
	def cb_menu_preferences(self, dragitem): 
		""" 
		callback from context menu
		opens preferences window 
		""" 
		import preferences_controller
		preferences_controller.shared.show_window()
	def cb_menu_quit(self, dragitem):
		""" 
		callback from context menu
		Quits the application 
		"""  
		from utils import quit_dragbox 
		quit_dragbox()
	def _remove_dragitem_if_gone(self, dragitem): 
		exists = dragitem.still_exists()
		if not exists:
			self._remove_dragitem(dragitem)
		return False
				 	 
	def _remove_dragitem(self, dragitem): 
		""" 
		Finds the child box of the window that
		contains the dragitem and removes it 
		""" 
		try:
			child = self.dragitems.pop(dragitem)
			self._remove_child(child)
		except KeyError:
			pass
	def _remove_child(self, child):
		""" 
		Removes a child of the main container widget
		@param child: the hbox child to be removed 
		@type child: a widget object.   
		 
		""" 
		self.vbox.remove(child)
		self.adjust_window_size(True)
	def cb_stick_pref_changed(client, connection, entry, user_info):
		""" 
		Callback from stick preference change
		
		@param client: gconf_client 
		@type client: an object. 
		
		@param connection:   
		@type connection: an object.  
		
		@param entry: The gconf entry 
		@type entry: a string object.    
		""" 
		from utils import print_debug
		print_debug("In window sticky callback") 
		self.window.set_sticky(client.get_bool(entry))

def make_menu(delegate, user_info):
	""" 
	Makes a context menu for a dragitem
	
	@param delegate: The object to receive menu callbacks 
	@type delegate: an object. 
	
	@param user_info: Passed as user_info with the callbacks
	@type user_info: an object.    
	""" 
	

	# menu items.
	items = [
		(gtk.STOCK_COPY, delegate.cb_menu_copy),
		(gtk.STOCK_CUT, delegate.cb_menu_cut),
		(gtk.STOCK_DELETE, delegate.cb_menu_remove),
		(None, None),		# Separator
		(gtk.STOCK_PREFERENCES, delegate.cb_menu_preferences),
		(gtk.STOCK_QUIT, delegate.cb_menu_quit)
		]
	return _make_menu(items, user_info)

def make_window_menu(delegate, user_info):
	""" 
	Makes a context menu for a the window
	
	@param delegate: The object to receive menu callbacks 
	@type delegate: an object. 
	
	@param user_info: Passed as user_info with the callbacks
	@type user_info: an object.    
	""" 
	# menu items.
	items = [
		(gtk.STOCK_PREFERENCES, delegate.cb_menu_preferences),
		(gtk.STOCK_QUIT, delegate.cb_menu_quit)
		]
	return _make_menu(items, user_info)

def _make_menu(items, user_info):
	cmenu = gtk.Menu()    # Don't need to show menus
	# Create the menu items
	for (name, callback) in items:
		menu_item = gtk.ImageMenuItem(name)
		cmenu.append(menu_item)
		if callback:
			menu_item.connect_object ("activate", callback, user_info)
	
	cmenu.show_all()
	return cmenu
